// Con Clase Enero 2024
// Ejemplo de lectura de rutas de fichero XML
// Añadir en las opciones de compilador:
// xml2.dll para usar la versión dinámica de libxml2

#include <iostream>
#include <string>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <windows.h>

void Evalua(const xmlChar *ruta, xmlXPathContextPtr xpathCtx);
void MostrarNodos(xmlNodeSetPtr nodos);
void MostrarHijos(xmlNodePtr node, int nivel);
void MostrarAtributos(xmlAttrPtr prop);

int main()
{
    xmlDocPtr doc;
    xmlXPathContextPtr xpathCtx;

    xmlInitParser();
    LIBXML_TEST_VERSION
    SetConsoleOutputCP(CP_UTF8);

    doc = xmlReadFile("ejemplo.xml", NULL, XML_PARSE_NOBLANKS);

    xpathCtx = xmlXPathNewContext(doc);

    Evalua((const xmlChar*)"/cursos/curso[2]/capitulo/nombre", xpathCtx);
    Evalua((const xmlChar*)"/cursos/curso/*", xpathCtx);
    Evalua((const xmlChar*)"/cursos/curso/capitulo/nombre/text()", xpathCtx);
    Evalua((const xmlChar*)"//@lang", xpathCtx);
    Evalua((const xmlChar*)"//capitulo[@orden='1']/nombre", xpathCtx);
    Evalua((const xmlChar*)"//capitulo[contains(nombre,'de')]/nombre", xpathCtx);
    Evalua((const xmlChar*)"child::cursos//capitulo", xpathCtx);
    Evalua((const xmlChar*)"//capitulo[numero='0']/nombre", xpathCtx);
    Evalua((const xmlChar*)"//capitulo[last()]/nombre", xpathCtx);
    Evalua((const xmlChar*)"//capitulo[position()<4]/nombre", xpathCtx);
    Evalua((const xmlChar*)"//capitulo[contains(nombre, ' ') or contains(nombre, '/')]/nombre", xpathCtx);
    Evalua((const xmlChar*)"//@lang|//nombre", xpathCtx);
    Evalua((const xmlChar*)"/cursos/curso[2]/capitulo[last()-1]/nombre", xpathCtx);
    Evalua((const xmlChar*)"/cursos/curso[1]/capitulo[position()=2]/nombre", xpathCtx);
    Evalua((const xmlChar*)"/cursos/curso[count(capitulo)=2]//nombre", xpathCtx);
    Evalua((const xmlChar*)"/cursos/curso[count(*)>2]//nombre", xpathCtx);
    Evalua((const xmlChar*)"/cursos//curso[substring(titulo, 1, 3)='Cur']//titulo", xpathCtx);
    Evalua((const xmlChar*)"/cursos//curso[concat('C','u','r','so')=substring(titulo, 1, 5)]//titulo", xpathCtx);

    std::cout << ">>>>>>>> Contexto en comentario" << std::endl;
    xmlXPathSetContextNode(doc->children->children, xpathCtx);
    Evalua((const xmlChar*)".", xpathCtx);

    std::cout << ">>>>>>>> Contexto en curso 1 capitulo 1" << std::endl;
    xmlXPathSetContextNode(doc->children->children->next->children->next, xpathCtx);
    Evalua((const xmlChar*)"../titulo", xpathCtx);
    Evalua((const xmlChar*)"//*", xpathCtx);
    Evalua((const xmlChar*)".", xpathCtx);
    Evalua((const xmlChar*)"@*", xpathCtx);

    /* Cleanup */
    xmlXPathFreeContext(xpathCtx);
    xmlFreeDoc(doc);
    xmlCleanupParser();
    return 0;
}

void Evalua(const xmlChar *ruta, xmlXPathContextPtr xpathCtx) {
    xmlXPathObjectPtr xpathObj;

    xpathObj = xmlXPathEvalExpression(ruta, xpathCtx);
    std::cout << "Path: " << ruta << std::endl;
    MostrarNodos(xpathObj->nodesetval);
    xmlXPathFreeObject(xpathObj);
}

void MostrarNodos(xmlNodeSetPtr nodos) {
    xmlNodePtr cur;
    int size;
    int i;

    size = (nodos) ? nodos->nodeNr : 0;

    std::cout << "Encontrados: " << size << " nodos" << std::endl;
    for(i = 0; i < size; ++i) {
        if(nodos->nodeTab[i]->type == XML_ELEMENT_NODE) {
            cur = nodos->nodeTab[i];
            std::cout << i+1 << ")  nodo elemento " << cur->name << " -> ";
            if(cur->children->type == XML_ELEMENT_NODE) {
                std::cout << std::endl;
                MostrarHijos(cur->children, 1);
            } else {
                std::cout << cur->children->content;
            }
            std::cout << std::endl;
        } else if(nodos->nodeTab[i]->type == XML_ATTRIBUTE_NODE) {
            cur = nodos->nodeTab[i];
            std::cout << i+1 << ")  nodo atributo " << cur->name << " -> ";
            if(cur->children) std::cout << cur->children->content;
            std::cout << std::endl;
        } else if(nodos->nodeTab[i]->type == XML_COMMENT_NODE) {
            cur = nodos->nodeTab[i];
            std::cout << i+1 << ")  nodo comentario " << cur->name << " -> ";
            if(cur->content) std::cout << cur->content;
            std::cout << std::endl;
        } else {
            cur = nodos->nodeTab[i];
            std::cout << i+1 << ")  nodo: " << cur->name << " Tipo: " << cur->type << std::endl;
        }
    }
}

void MostrarHijos(xmlNodePtr node, int nivel) {
    while(node) {
        if(node->type == XML_ELEMENT_NODE) {
            std::cout << std::endl;
            for(int i=0; i<nivel; i++) std::cout << ' ';
            std::cout << "Elemento: " << node->name << " - ";
            MostrarAtributos(node->properties);
        } else
        if(node->type == XML_TEXT_NODE || node->type == XML_COMMENT_NODE) { // Comentarios o texto
            if(!xmlIsBlankNode(node)) {
                std::cout << " Nodo tipo(" << node->type << ") " << node->name << " - " << node->content;
            }
        }
        MostrarHijos(node->children, nivel+1);
        node = node->next;
    }
}

void MostrarAtributos(xmlAttrPtr prop) {
    while(prop) {
        if(prop->type == XML_ATTRIBUTE_NODE) {
            std::cout << " " << prop->name;
            std::cout << "(" << prop->children->content << ") ";
        }
        prop = prop->next;
    }
}
